package classifier;

import java.util.*;
import classifier.calculators.*;
import classifier.knowledge_tree.*;
import converter.*;

public class Knn
{

	public Knn(int neighbourhoods)
	{
		k = neighbourhoods;
		is_trained = false;
		knowledge_tree = null;
		net = new Net();
		entropy_plus_diversity = false;
		use_lists = true;
	}
	
	//stores the train instances
	//code:
	/*
	0: Overlap
	1: Modified Value
	*/
	public void train(ArrayList<Instance> knowledge, int code)
	{
		//empty train set
		if(knowledge.size() == 0)
		{
			return;
		}
		net.train(knowledge);
		//create a notion for the ambitag of a word or an ending of a word
		//according to the train set
		Word_and_ending_ambitag_calculator ambitags = new Word_and_ending_ambitag_calculator();
		ambitags.calculate_word_and_ending_ambitag(knowledge);
		word_ambitag = ambitags.get_word_ambitags();
		ending_ambitag = ambitags.get_ending_ambitags();
		//create an alternative representation - use of overlap metric by default
		knowledge_tree = new Knowledge_tree(knowledge, code);
		is_trained = true;
	}
	
	//sets whether the classifier uses lists or not
	public void use_lists(boolean lists)
	{
		use_lists = lists;
	}
	
	//sets the selection measure
	public void set_selection_measure(boolean e_p_d)
	{
		if(!is_trained)
		{
			throw new IllegalStateException("First train then set selection measure.");
		}
		entropy_plus_diversity = e_p_d;
	}
	
	//apply different weight to different properties
	public void use_properties_weight(boolean b)
	{
		if(!is_trained)
		{
			throw new IllegalStateException("First train then set weight parameter.");
		}
		knowledge_tree.use_properties_weight(b);
	}
	
	//return the ambitag of the given word according to what can be concluded given
	//the train set
	public String get_ambitag(String word)
	{
		String ambitag = word_ambitag.get(word);
		if(ambitag == null)
		{
			ambitag = ending_ambitag.get(get_ending(word));
		}
		if(ambitag != null)
		{
			return ambitag;
		}
		return "unknown";
	}
	
	//returns the ending of a word
	private String get_ending(String word)
	{
		if(word.equals("null"))
		{
			return "null";
		}
		//if the word is less than 3 characters long
		if(word.length()<=3)
		{//its ending is the word itself
			return word;
		}
		//else return the three last characters
		return word.substring(
				word.length()-3,
				word.length());
	}
	
	//return the category of an incoming instance in
	public String classify(Instance in)
	{
		if(!is_trained)
		{
			throw new IllegalStateException("Classifier is not trained therefore cannot " +
					"classify.");
		}
		if(use_lists)
		{
			String category = net.classify(in);
			if(category != null)
			{
				return category;
			}
		}
		//the knowledge_tree using DFS and pruning will find the k nearest neighbourhoods fast
		//get all neighbours from the k closest neighbourhoods, each votes for its category
		return Converter.reverse(
				get_most_common_category(knowledge_tree.get_k_nearest_neighbourhoods(k, in)));
	}
	
	//returns whether this instance's word is contained in a list or not
	public boolean is_contained_in_list(Instance in)
	{
		if(net.classify(in) != null)
		{
			return true;
		}
		return false;
	}
	
	//returns the uncertainty of the classifier for the given instance
	public double get_uncertainty_for_classification(Instance in)
	{
		if(!is_trained)
		{
			throw new IllegalStateException("Classifier is not trained therefore cannot " +
			"calculate uncertainty.");
		}
		TreeMap<Double, Neighbourhood> k_nearest_neighbourhoods_sorted =
			knowledge_tree.get_k_nearest_neighbourhoods(k, in);
		//keep in a map the votes for each category
		TreeMap<String, Double> category_votes = new TreeMap<String, Double>();
		//for each neighbourhood
		Iterator<Double> it = k_nearest_neighbourhoods_sorted.keySet().iterator();
		while(it.hasNext())
		{
			Double distance = it.next();
			ArrayList<String> neighbours =
				k_nearest_neighbourhoods_sorted.get(distance).get_neighbours();
			//for every neighbour
			for(int j=0; j<neighbours.size(); ++j)
			{
				String instance_category = neighbours.get(j);
				//distance-weighted class voting
				//constant chosen 1.0E-10
				double instance_vote = 1/
				(distance.doubleValue()+1.0E-10);
				//add its vote to the corresponding category
				if(category_votes.containsKey(instance_category))
				{//category already has a vote
					Double new_value = new Double(
					category_votes.get(instance_category).doubleValue() + 
					instance_vote);
					category_votes.put(instance_category, new_value);
				}
				else
				{//new category found
					category_votes.put(instance_category, instance_vote);
				}
			}
		}
		
		//find the sum of the votes of all the categories
		double vote_of_all_categories = 0;
		Iterator<String> it2 = category_votes.keySet().iterator();
		while(it2.hasNext())
		{
			vote_of_all_categories += category_votes.get(it2.next());
		}
		//find the possibility for this instance to be of any category and find its entropy
		double sum = 0;
		Iterator<String> it3 = category_votes.keySet().iterator();
		while(it3.hasNext())
		{
			double category_possibility = category_votes.get(it3.next())/vote_of_all_categories;
			sum += category_possibility*(Math.log(category_possibility)/Math.log(2));
		}
		//-sum here is the H(x) in the paper
		double normalized_entropy;
		if(category_votes.size() == 1)
		{
			normalized_entropy = 0;
		}
		else
		{
			normalized_entropy = -sum/(Math.log(category_votes.size())/Math.log(2));
		}
		if(normalized_entropy == 0)
		{
			return -vote_of_all_categories;
		}
		if(entropy_plus_diversity)
		{
			return normalized_entropy/(Math.log(vote_of_all_categories)/Math.log(2));
		}
		return normalized_entropy;
	}
	
	//returns the category with the highest overall votes
	private String get_most_common_category(TreeMap<Double, Neighbourhood>
	                                                k_nearest_neighbourhoods_sorted)
	{
		//keep in a map the votes for each category
		TreeMap<String, Double> category_votes = new TreeMap<String, Double>();
		//keep in a map the times each category is found
		TreeMap<String, Integer> category_times = new TreeMap<String, Integer>();
		//for each neighbourhood
		Iterator<Double> it = k_nearest_neighbourhoods_sorted.keySet().iterator();
		while(it.hasNext())
		{
			Double distance = it.next();
			ArrayList<String> neighbours =
				k_nearest_neighbourhoods_sorted.get(distance).get_neighbours();
			//for every neighbour
			for(int j=0; j<neighbours.size(); ++j)
			{
				String instance_category = neighbours.get(j);
				//distance-weighted class voting
				//constant chosen 1.0E-10
				double instance_vote = 1/
				(distance.doubleValue()+1.0E-10);
				//add its vote to the corresponding category
				if(category_votes.containsKey(instance_category))
				{//category already has a vote
					Double new_value = new Double(
					category_votes.get(instance_category).doubleValue() + 
					instance_vote);
					category_votes.put(instance_category, new_value);
					//also update category_times map
					category_times.put(instance_category, new Integer(category_times.get(
							instance_category).intValue()+1));
				}
				else
				{//new category found
					category_votes.put(instance_category, instance_vote);
					//also put in the category_times map
					category_times.put(instance_category, new Integer(1));
				}
			}
		}
		//tie braking
		return brake_tie(get_categories_with_highest_vote(category_votes), category_times);
	}

	//locates the categories with the highest vote
	private ArrayList<String> get_categories_with_highest_vote(
			TreeMap<String, Double> category_votes)
	{
		Iterator<String> it = category_votes.keySet().iterator();
		double max = 0;
		//here we hold all the categories that have the highest vote
		ArrayList<String> categories_with_highest_vote = new ArrayList<String>();
		//for each category
		while(it.hasNext())
		{
			String category = it.next();
			//get its overall vote
			Double temp = category_votes.get(category);
			if(temp.doubleValue() == max)
			{//keep the category
				categories_with_highest_vote.add(category);
			}
			if(temp.doubleValue() > max)
			{//keep the highest vote
				max = temp.doubleValue();
				//new highest vote found, clear the list with the highest votes
				categories_with_highest_vote = new ArrayList<String>();
				//keep only the new category
				categories_with_highest_vote.add(category);
			}
		}
		return categories_with_highest_vote;
	}
	
	//if more than one categories have the highest vote then we randomly choose one
	//the possibility of a category to be chosen is set according to the times this category
	//was found in the train instances that voted
	private String brake_tie(ArrayList<String> categories_with_highest_vote,
			TreeMap<String, Integer> category_times)
	{
		int sum_of_times_for_all_high_categories = 0;
		for(int i=0; i<categories_with_highest_vote.size(); ++i)
		{
			sum_of_times_for_all_high_categories += category_times.get(
					categories_with_highest_vote.get(i));
		}
		int random_number = (int)(Math.random()*sum_of_times_for_all_high_categories);
		for(int i=0; i<categories_with_highest_vote.size(); ++i)
		{
			random_number -= category_times.get(categories_with_highest_vote.get(i));
			if(random_number < 0)
			{
				return categories_with_highest_vote.get(i);
			}
		}
		//unreachable code
		return "";
	}
	
	//the number of neighbourhoods we check
	private int k;
	//a flag to mark whether the classifier is trained or not
	private boolean is_trained;
	//the classifier's notion for the ambitag of a word
	private HashMap<String, String> word_ambitag;
	//the classifier's notion for the ambitag of an ending of a word
	private HashMap<String, String> ending_ambitag;
	//a different representation of the train set that allows faster classifications
	public Knowledge_tree knowledge_tree;
	//a middleware were some tokens are automatically tagged
	private Net net;
	//the measure to calculate uncertainty for classification
	private boolean entropy_plus_diversity;
	//use the Net or not
	private boolean use_lists;
	
}
