package classifier.knowledge_tree;

import classifier.*;
import classifier.calculators.Gain_ratio_calculator;
import classifier.property_value_difference_metric.*;
import java.util.*;

public class Knowledge_tree
{
	
	//given an ArrayList with instances it constructs the corresponding knowledge_tree
	public Knowledge_tree(ArrayList<Instance> knowledge_in, int code)
	{
		//find the gain ratio of each property
		properties_weight = new Gain_ratio_calculator().get_properties_weight(knowledge_in);
		//by default, don't use different weight for each property 
		gain_ratio = false;
		if(code == 0)
		{
			knowledge = knowledge_in;
			root = null;
			difference_metric = new Overlap_difference_metric(knowledge);
		}
		else
		{
			knowledge = null;
			difference_metric = new Modified_value_difference_metric(knowledge_in);
			//the root is of depth 0
			root = new Knowledge_node((byte)0);
			//adds all instances to the root - they are automatically distributed in the whole tree
			//and finally all instances go to the leafs
			while(!knowledge_in.isEmpty())
			{
				root.add_instance(knowledge_in.remove(0));
			}
		}
	}
	
	//apply different weight to different properties
	public void use_properties_weight(boolean b)
	{
		gain_ratio = b;
	}
	
	//finds the k nearest neighbourhoods of the incoming instance
	public TreeMap<Double, Neighbourhood> get_k_nearest_neighbourhoods(int k, Instance in)
	{
		//initiate k closest distances - neighbourhoods.
		TreeMap<Double, Neighbourhood> k_nearest_neighbourhoods_sorted =
			new TreeMap<Double, Neighbourhood>();
		for(int i=0; i<k; ++i)
		{
			//maximum distance at first
			k_nearest_neighbourhoods_sorted.
			put(new Double(Double.MAX_VALUE), new Neighbourhood(Double.MAX_VALUE));
		}
		//if the difference metric is the modified_value_difference_metric
		//we use the tree
		if(difference_metric instanceof Modified_value_difference_metric)
		{
			//parse the tree and updates the k_nearest_neighbourhoods structure
			DFS(root, 0, k_nearest_neighbourhoods_sorted, in);
		}
		//else we use the array list
		else
		{
			parse_array(k_nearest_neighbourhoods_sorted, in);
		}
		return k_nearest_neighbourhoods_sorted;
	}
	
	//parses the tree using the DFS algorithm and prunes when able to
	private void DFS(Knowledge_node node, double distance,
			TreeMap<Double, Neighbourhood> knns, Instance in)
	{
		//if the distance so far is larger than the distance of the farthest neighbourhood
		//don't look any further under this node - prune
		if(distance > knns.lastKey())
		{
			return;
		}
		//if node is a leaf - depth 15, all properties have been checked
		if(node.get_depth() == in.get_number_of_properties())
		{
			//if the distance found is smaller than the distance of the farthest
			//neighbourhood construct a new neighbourhood and discard the farthest one
			if(distance < knns.lastKey())
			{
				knns.pollLastEntry();
				Neighbourhood new_neighbourhood = new Neighbourhood(distance);
				//add all instances with these property values
				for(int i=0; i<node.get_instances().size(); ++i)
				{
					new_neighbourhood.add_neighbour(node.get_instances().get(i));
				}
				knns.put(new Double(distance), new_neighbourhood);
				return;
			}
			//if a neighbourhood with this distance already exists
			//log(n) for the containsKey operation
			if(knns.containsKey(new Double(distance)))
			{
				//add all instances with these property values
				for(int i=0; i<node.get_instances().size(); ++i)
				{
					knns.get(new Double(distance)).add_neighbour(node.get_instances().get(i));
				}
				return;
			}
		}
		//for each edge that comes out of the given node (each edge corresponds to a different value
		//of the property whose index in the instance can be obtained by the node's depth)
		Iterator<String> property_values = node.get_children().keySet().iterator();
		String current_value = "";
		while(property_values.hasNext())
		{
			//get the property value
			current_value = property_values.next();
			//construct a phantom instance to feed the difference_metric
			Instance temp_in = new Instance();
			//the target property must be set to have the right value 
			temp_in.set_property_value(node.get_depth(), current_value);
			double values_difference = 
				difference_metric.get_difference(node.get_depth(), temp_in, in);
			//the weight of the property
			if(gain_ratio)
			{
				values_difference = values_difference*properties_weight[node.get_depth()];
			}
			//1st parameter: the node who is the child under the current edge
			//2nd parameter: the distance so far plus the properties' values difference
			//biased by the corresponding property weight (if selected)
			DFS(node.get_children().get(current_value),
					distance+values_difference,
					knns, in);
		}
		
	}
	
	//finds the k nearest neighbourhoods using the array structure
	private void parse_array(TreeMap<Double, Neighbourhood>
											k_nearest_neighbourhoods_sorted, Instance in)
	{
		//check distance from each vector learned in the training phase
		double distance;
		Instance temp;
		for(int i=0; i<knowledge.size(); ++i)
		{
			temp = new Instance(knowledge.get(i));
			distance = distance(temp, in);
			//rearrange the closest ones if needed
			//if the new temp instance is at least as close as the farthest neighbourhood
			if(distance <= k_nearest_neighbourhoods_sorted.lastKey())
			{
				update_neighbourhoods(distance, temp, k_nearest_neighbourhoods_sorted);
			}
		}
	}
	
	//returns the distance between instance a and instance b
	private double distance(Instance a, Instance b)
	{
		double distance = 0;
		double property_difference = 0;
		double property_weight = 1;
		for(int i=0; i<a.get_number_of_properties(); ++i)
		{
			property_difference = difference_metric.get_difference(i, a, b);
			if(gain_ratio)
			{
				property_weight = properties_weight[i];
			}
			distance += property_weight*property_difference;
		}
		return distance;
	}
	
	//insert new_in instance with new_dist distance to the corresponding neighbourhood
	private void update_neighbourhoods(double new_dist, Instance new_in,
				TreeMap<Double, Neighbourhood> k_nearest_neighbourhoods_sorted)
	{
		//locate the right neighbourhood for the new instance
		Neighbourhood neighbourhood = k_nearest_neighbourhoods_sorted.get(new_dist);
		//if no neighbourhood exists for the given distance
		if(neighbourhood == null)
		{
			//create a new one
			neighbourhood = new Neighbourhood(new_dist);
			//discard the farthest one
			k_nearest_neighbourhoods_sorted.pollLastEntry();
			//and add the new neighbourhood to the map
			k_nearest_neighbourhoods_sorted.put(new_dist, neighbourhood);
		}
		//finally, add the incoming instance to the corresponding neighbourhood
		neighbourhood.add_neighbour(new_in.get_category());
	}
	
	//the root of the tree
	private Knowledge_node root;
	//a way to measure the difference between two values of a property
	private Property_value_difference_metric difference_metric;
	//the instances in an array
	private ArrayList<Instance> knowledge;
	//the weight of each property
	private double[] properties_weight;
	//a flag to determine whether we use different weight for each property or not
	private boolean gain_ratio;
	
}
