package document;

import java.io.Serializable;
import java.util.*;

public class Token_document implements Serializable
{

	public Token_document()
	{
		tokens = new TreeMap<Integer, Token>();
		sentences = new TreeMap<Integer, Token_sentence>();
		true_token_positions = new HashMap<Token, Integer[]>();
		normalized_token_positions = new HashMap<Token, Integer[]>();
	}
	
	//adds a sentence in the document
	public void add_sentence(Token_sentence sentence)
	{
		if(sentences.get(sentence.get_first_token().get_starting_position())
				!= null)
		{
			//another sentence with the same starting position exists - illegal in the same document
			throw new IllegalArgumentException("Two sentences with the same starting position" +
					" in the same document!");
		}
		sentences.put(sentence.get_first_token().get_starting_position(), sentence);
	}
	
	//adds a token in the document
	public void add_token(Token token)
	{
		if(tokens.get(token.get_starting_position()) != null)
		{
			//another token with the same starting position exists - illegal in the same document
			throw new IllegalArgumentException("Two tokens with the same starting position" +
					" in the same document!");
		}
		tokens.put(token.get_starting_position(), token);
		true_token_positions.put(token, new Integer[]{token.get_starting_position(),
				token.get_ending_position()});
	}
	
	//returns the token that is hit if the cursor is in the given cursor_position position
	//if any exists - if non exist return null
	public Token get_token_in_cursor_position(int cursor_position)
	{
		//go to previous cursor position until a token's starting position is found
		int mark = cursor_position;
		while(tokens.firstKey() <= mark)
		{
			if(tokens.get(mark) == null)
			{
				mark--;
				continue;
			}
			return tokens.get(mark);
		}
		return tokens.firstEntry().getValue();
	}
	
	//returns all the tokens in a collection
	public Collection<Token> get_tokens()
	{
		return tokens.values();
	}
	
	//returns the previous token
	public Token get_next_token(Token token)
	{
		//the incoming token is the highest
		if(tokens.lastKey() == token.get_starting_position())
		{
			return token;
		}
		//go to next cursor position until a token's starting position is found
		int mark = token.get_ending_position();
		while(tokens.lastKey() >= mark)
		{
			if(tokens.get(mark) == null)
			{
				mark++;
				continue;
			}
			return tokens.get(mark);
		}
		//unreachable code
		return null;
	}
	
	//returns the next token
	public Token get_previous_token(Token token)
	{
		//the incoming token is the lowest
		if(tokens.firstKey() == token.get_starting_position())
		{
			return token;
		}
		//go to previous cursor position until a token's starting position is found
		int mark = token.get_starting_position()-1;
		while(tokens.firstKey() <= mark)
		{
			if(tokens.get(mark) == null)
			{
				mark--;
				continue;
			}
			return tokens.get(mark);
		}
		//unreachable code
		return null;
	}
	
	//returns the first sentence of the document
	public Token_sentence get_first_sentence()
	{
		if(sentences.firstEntry() != null)
		{
			return sentences.firstEntry().getValue();
		}
		return null;
	}
	
	//returns the next sentence of the given one
	public Token_sentence get_next_sentence(Token_sentence sentence)
	{
		//if this is the last sentence return null
		if(sentences.lastEntry().getValue() == sentence)
		{
			return null;
		}
		//else return the next sentence
		return sentences.get(get_next_token(sentence.get_last_token()).get_starting_position());
	}
	
	//returns the sentence of the given token
	public Token_sentence get_sentence_of_token(Token token)
	{
		int mark = token.get_starting_position();
		while(sentences.get(mark) == null)
		{
			mark--;
		}
		return sentences.get(mark);
	}
	
	//changes the format of the document
	public void normalize_document(boolean normalize)
	{
		//if the normalization is requested for the first time
		if(!(normalized_token_positions.size() == true_token_positions.size()))
		{
			//get all the tokens
			Iterator<Integer> it = tokens.keySet().iterator();
			int normalized_starting_position = 0;
			while(it.hasNext())
			{
				Token temp = tokens.get(it.next());
				normalized_token_positions.put(temp, new Integer[]{
						normalized_starting_position,
						normalized_starting_position+temp.get_token().length()});
				normalized_starting_position += temp.get_token().length()+1;
			}
		}
		//if the normalized document is requested
		if(normalize)
		{
			//update all the tokens
			Iterator<Integer> it = tokens.keySet().iterator();
			while(it.hasNext())
			{
				Token temp = tokens.get(it.next());
				temp.set_starting_position(normalized_token_positions.get(temp)[0]);
				temp.set_ending_position(normalized_token_positions.get(temp)[1]);
			}
		}
		//else the true document is requested
		else
		{
			//update all the tokens
			Iterator<Integer> it = tokens.keySet().iterator();
			while(it.hasNext())
			{
				Token temp = tokens.get(it.next());
				temp.set_starting_position(true_token_positions.get(temp)[0]);
				temp.set_ending_position(true_token_positions.get(temp)[1]);
			}
		}
		//update the tokens structure
		TreeMap<Integer, Token> temp = new TreeMap<Integer, Token>();
		Iterator<Integer> it = tokens.keySet().iterator();
		while(it.hasNext())
		{
			Token token = tokens.get(it.next());
			temp.put(token.get_starting_position(), token);
		}
		tokens = temp;
		//update the sentences structure
		TreeMap<Integer, Token_sentence> temp2 = new TreeMap<Integer, Token_sentence>();
		Iterator<Integer> it2 = sentences.keySet().iterator();
		while(it2.hasNext())
		{
			Token_sentence sentence = sentences.get(it2.next());
			temp2.put(sentence.get_first_token().get_starting_position(), sentence);
		}
		sentences = temp2;
	}
	
	//returns the normalized format of the text
	public String get_normalized_text()
	{
		String normalized_text = "";
		Token_sentence sentence = get_first_sentence();
		//for each sentence
		while(sentence != null)
		{
			Token current = sentence.get_first_token();
			Token last = sentence.get_last_token();
			//for each token in the sentence
			while(current != last)
			{
				normalized_text += current.get_token() + " ";
				current = get_next_token(current);
			}
			normalized_text += last.get_token() + "\n";
			sentence = get_next_sentence(sentence);
		}
		return normalized_text;
	}
	
	//this map holds the (different) true starting positions of the tokens and the tokens
	private TreeMap<Integer, Token> tokens;
	//this map holds the (different) true starting positions of the sentences and the sentences
	private TreeMap<Integer, Token_sentence> sentences;
	private HashMap<Token, Integer[]> true_token_positions;
	private HashMap<Token, Integer[]> normalized_token_positions;
	
}










